@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -112,20 +112,33 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 	BIO *dtls_bio;
 	int dtls_fd;
 
+	if (!vpninfo->dtls_addr) {
+		vpninfo->progress(vpninfo, PRG_ERR, "No DTLS address\n");
+		vpninfo->dtls_attempt_period = 0;
+		return -EINVAL;
+	}
+
 	if (!vpninfo->dtls_cipher) {
 		/* We probably didn't offer it any ciphers it liked */
 		vpninfo->progress(vpninfo, PRG_ERR, "Server offered no DTLS cipher option\n");
 		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
-		
+
+	if (vpninfo->proxy) {
+		/* XXX: Theoretically, SOCKS5 proxies can do UDP too */
+		vpninfo->progress(vpninfo, PRG_ERR, "No DTLS when connected via proxy\n");
+		vpninfo->dtls_attempt_period = 0;
+		return -EINVAL;
+	}
+
 	dtls_fd = socket(vpninfo->peer_addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
 	if (dtls_fd < 0) {
 		perror("Open UDP socket for DTLS:");
 		return -EINVAL;
 	}
 
-	if (connect(dtls_fd, vpninfo->peer_addr, vpninfo->peer_addrlen)) {
+	if (connect(dtls_fd, vpninfo->dtls_addr, vpninfo->peer_addrlen)) {
 		perror("UDP (DTLS) connect:\n");
 		close(dtls_fd);
 		return -EINVAL;
@@ -329,6 +342,7 @@ int setup_dtls(struct openconnect_info *vpninfo)
 			if (strlen(dtls_opt->value) != 64) {
 				vpninfo->progress(vpninfo, PRG_ERR, "X-DTLS-Session-ID not 64 characters\n");
 				vpninfo->progress(vpninfo, PRG_ERR, "Is: %s\n", dtls_opt->value);
+				vpninfo->dtls_attempt_period = 0;
 				return -EINVAL;
 			}
 			for (i = 0; i < 64; i += 2)
@@ -348,21 +362,30 @@ int setup_dtls(struct openconnect_info *vpninfo)
 
 		dtls_opt = dtls_opt->next;
 	}
-	if (!sessid_found || !dtls_port)
-		return -EINVAL;
-
-	if (vpninfo->peer_addr->sa_family == AF_INET) {
-		struct sockaddr_in *sin = (void *)vpninfo->peer_addr;
-		sin->sin_port = htons(dtls_port);
-	} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
-		struct sockaddr_in6 *sin = (void *)vpninfo->peer_addr;
-		sin->sin6_port = htons(dtls_port);
-	} else {
-		vpninfo->progress(vpninfo, PRG_ERR, "Unknown protocol family %d. Cannot do DTLS\n",
-			vpninfo->peer_addr->sa_family);
+	if (!sessid_found || !dtls_port) {
+		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
 
+	vpninfo->dtls_addr = malloc(vpninfo->peer_addrlen);
+	if (!vpninfo->dtls_addr) {
+		vpninfo->dtls_attempt_period = 0;
+		return -ENOMEM;
+	}
+	memcpy(vpninfo->dtls_addr, vpninfo->peer_addr, vpninfo->peer_addrlen);
+
+	if (vpninfo->peer_addr->sa_family == AF_INET) {
+		struct sockaddr_in *sin = (void *)vpninfo->dtls_addr;
+		sin->sin_port = htons(dtls_port);
+	} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
+		struct sockaddr_in6 *sin = (void *)vpninfo->dtls_addr;
+		sin->sin6_port = htons(dtls_port);
+	} else {
+		vpninfo->progress(vpninfo, PRG_ERR, "Unknown protocol family %d. Cannot do DTLS\n",
+				  vpninfo->peer_addr->sa_family);
+		vpninfo->dtls_attempt_period = 0;
+		return -EINVAL;
+	}
 
 	if (connect_dtls_socket(vpninfo))
 		return -EINVAL;
