@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -40,6 +40,8 @@
 
 #include "openconnect.h"
 
+static int proxy_write(int fd, unsigned char *buf, size_t len);
+
 #define MAX_BUF_LEN 131072
 /*
  * We didn't really want to have to do this for ourselves -- one might have
@@ -94,28 +96,30 @@ int http_add_cookie(struct openconnect_info *vpninfo, const char *option, const
 	return 0;
 }
 
+#define BODY_HTTP10 -1
+#define BODY_CHUNKED -2
+
 static int process_http_response(struct openconnect_info *vpninfo, int *result,
 				 int (*header_cb)(struct openconnect_info *, char *, char *),
-				 char *body, int buf_len)
+				 char **body_ret)
 {
 	char buf[MAX_BUF_LEN];
-	int bodylen = 0;
+	char *body = NULL;
+	int bodylen = BODY_HTTP10;
 	int done = 0;
-	int http10 = 0, closeconn = 0;
+	int closeconn = 0;
 	int i;
 
+ cont:
 	if (openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)) < 0) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTPS response\n");
 		return -EINVAL;
 	}
 
- cont:
-	if (!strncmp(buf, "HTTP/1.0 ", 9)) {
-		http10 = 1;
+	if (!strncmp(buf, "HTTP/1.0 ", 9))
 		closeconn = 1;
-	}
-
-	if ((!http10 && strncmp(buf, "HTTP/1.1 ", 9)) || !(*result = atoi(buf+9))) {
+	
+	if ((!closeconn && strncmp(buf, "HTTP/1.1 ", 9)) || !(*result = atoi(buf+9))) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse HTTP response '%s'\n", buf);
 		return -EINVAL;
 	}
@@ -141,23 +145,33 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		if (*colon == ' ')
 			colon++;
 
-		if (!strcmp(buf, "Connection") && !strcmp(colon, "Close"))
-			closeconn = 1;
-
-		if (!strcmp(buf, "Location")) {
+		if (!strcasecmp(buf, "Connection")) {
+			if (!strcasecmp(colon, "Close"))
+				closeconn = 1;
+#if 0
+			/* This might seem reasonable, but in fact it breaks
+			   certificate authentication with some servers. If
+			   they give an HTTP/1.0 response, even if they
+			   explicitly give a Connection: Keep-Alive header,
+			   just close the connection. */
+			else if (!strcasecmp(colon, "Keep-Alive"))
+				closeconn = 0;
+#endif
+		}
+		if (!strcasecmp(buf, "Location")) {
 			vpninfo->redirect_url = strdup(colon);
 			if (!vpninfo->redirect_url)
 				return -ENOMEM;
 		}
-		if (!strcmp(buf, "Content-Length")) {
+		if (!strcasecmp(buf, "Content-Length")) {
 			bodylen = atoi(colon);
-			if (bodylen < 0 || bodylen > buf_len) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Response body too large for buffer (%d > %d)\n",
-					bodylen, buf_len);
+			if (bodylen < 0) {
+				vpninfo->progress(vpninfo, PRG_ERR, "Response body has negative size (%d)\n",
+						  bodylen);
 				return -EINVAL;
 			}
 		}
-		if (!strcmp(buf, "Set-Cookie")) {
+		if (!strcasecmp(buf, "Set-Cookie")) {
 			char *semicolon = strchr(colon, ';');
 			char *equals = strchr(colon, '=');
 			int ret;
@@ -175,9 +189,9 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			if (ret)
 				return ret;
 		}
-		if (!strcmp(buf, "Transfer-Encoding")) {
-			if (!strcmp(colon, "chunked"))
-				bodylen = -1;
+		if (!strcasecmp(buf, "Transfer-Encoding")) {
+			if (!strcasecmp(colon, "chunked"))
+				bodylen = BODY_CHUNKED;
 			else {
 				vpninfo->progress(vpninfo, PRG_ERR, "Unknown Transfer-Encoding: %s\n", colon);
 				return -EINVAL;
@@ -192,80 +206,99 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		goto cont;
 
 	/* Now the body, if there is one */
-	if (!bodylen)
-		goto fin;
+	vpninfo->progress(vpninfo, PRG_TRACE, "HTTP body %s (%d)\n", 
+			  bodylen==BODY_HTTP10?"http 1.0" :
+			  bodylen==BODY_CHUNKED?"chunked" : "length: ",
+			  bodylen);
 
-	if (http10) {
-		/* HTTP 1.0 response. Just eat all we can. */
-		while (1) {
-			i = SSL_read(vpninfo->https_ssl, body + done, bodylen - done);
-			if (i < 0)
-				goto fin;
-			done += i;
-		}
-	}
 	/* If we were given Content-Length, it's nice and easy... */
 	if (bodylen > 0) {
+		body = malloc(bodylen + 1);
+		if (!body)
+			return -ENOMEM;
 		while (done < bodylen) {
 			i = SSL_read(vpninfo->https_ssl, body + done, bodylen - done);
 			if (i < 0) {
 				vpninfo->progress(vpninfo, PRG_ERR, "Error reading HTTP response body\n");
+                                free(body);
 				return -EINVAL;
 			}
 			done += i;
 		}
-		goto fin;
-	}
+	} else if (bodylen == BODY_CHUNKED) {
+		/* ... else, chunked */
+		while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+			int chunklen, lastchunk = 0;
 
-	/* ... else, chunked */
-	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
-		int chunklen, lastchunk = 0;
-
-		if (i < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Error fetching chunk header\n");
-			exit(1);
-		}
-		chunklen = strtol(buf, NULL, 16);
-		if (!chunklen) {
-			lastchunk = 1;
-			goto skip;
-		}
-		if (chunklen + done > buf_len) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Response body too large for buffer (%d > %d)\n",
-				chunklen + done, buf_len);
-			return -EINVAL;
-		}
-		while (chunklen) {
-			i = SSL_read(vpninfo->https_ssl, body + done, chunklen);
 			if (i < 0) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Error reading HTTP response body\n");
+				vpninfo->progress(vpninfo, PRG_ERR, "Error fetching chunk header\n");
+				exit(1);
+			}
+			chunklen = strtol(buf, NULL, 16);
+			if (!chunklen) {
+				lastchunk = 1;
+				goto skip;
+			}
+			body = realloc(body, done + chunklen + 1);
+			if (!body)
+				return -ENOMEM;
+			while (chunklen) {
+				i = SSL_read(vpninfo->https_ssl, body + done, chunklen);
+				if (i < 0) {
+					vpninfo->progress(vpninfo, PRG_ERR, "Error reading HTTP response body\n");
+                                        free(body);
+					return -EINVAL;
+				}
+				chunklen -= i;
+				done += i;
+			}
+		skip:
+			if ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+				if (i < 0) {
+					vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTP response body\n");
+				} else {
+					vpninfo->progress(vpninfo, PRG_ERR, "Error in chunked decoding. Expected '', got: '%s'",
+							  buf);
+				}
+                                free(body);
 				return -EINVAL;
 			}
-			chunklen -= i;
-			done += i;
+
+			if (lastchunk)
+				break;
 		}
-	skip:
-		if ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
-			if (i < 0) {
-				vpninfo->progress(vpninfo, PRG_ERR, "Error fetching HTTP response body\n");
-			} else {
-				vpninfo->progress(vpninfo, PRG_ERR, "Error in chunked decoding. Expected '', got: '%s'",
-					buf);
-			}
+	} else if (bodylen == BODY_HTTP10) {
+		if (!closeconn) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Cannot receive HTTP 1.0 body without closing connection\n");
 			return -EINVAL;
 		}
 
-		if (lastchunk)
-			break;
+		/* HTTP 1.0 response. Just eat all we can in 16KiB chunks */
+		while (1) {
+			body = realloc(body, done + 16384);
+			if (!body)
+				return -ENOMEM;
+			i = SSL_read(vpninfo->https_ssl, body + done, 16384);
+			if (i <= 0) {
+				body = realloc(body, done + 1);
+				if (!body)
+					return -ENOMEM;
+				break;
+                        }
+			done += i;
+		}
 	}
- fin:
+
 	if (closeconn) {
 		SSL_free(vpninfo->https_ssl);
 		vpninfo->https_ssl = NULL;
 		close(vpninfo->ssl_fd);
 		vpninfo->ssl_fd = -1;
 	}
-	body[done] = 0;
+
+	if (body)
+		body[done] = 0;
+	*body_ret = body;
 	return done;
 }
 
@@ -274,6 +307,7 @@ static int fetch_config(struct openconnect_info *vpninfo, char *fu, char *bu,
 {
 	struct vpn_option *opt;
 	char buf[MAX_BUF_LEN];
+	char *config_buf = NULL;
 	int result, buflen;
 	unsigned char local_sha1_bin[SHA_DIGEST_LENGTH];
 	char local_sha1_ascii[(SHA_DIGEST_LENGTH * 2)+1];
@@ -296,18 +330,19 @@ static int fetch_config(struct openconnect_info *vpninfo, char *fu, char *bu,
 
 	SSL_write(vpninfo->https_ssl, buf, strlen(buf));
 
-	buflen = process_http_response(vpninfo, &result, NULL, buf, MAX_BUF_LEN);
+	buflen = process_http_response(vpninfo, &result, NULL, &config_buf);
 	if (buflen < 0) {
 		/* We'll already have complained about whatever offended us */
 		return -EINVAL;
 	}
 
-	if (result != 200)
+	if (result != 200) {
+		free(config_buf);
 		return -EINVAL;
-
+	}
 
 	EVP_MD_CTX_init(&c);
-	EVP_Digest(buf, buflen, local_sha1_bin, NULL, EVP_sha1(), NULL);
+	EVP_Digest(config_buf, buflen, local_sha1_bin, NULL, EVP_sha1(), NULL);
 	EVP_MD_CTX_cleanup(&c);
 
 	for (i = 0; i < SHA_DIGEST_LENGTH; i++)
@@ -315,16 +350,19 @@ static int fetch_config(struct openconnect_info *vpninfo, char *fu, char *bu,
 
 	if (strcasecmp(server_sha1, local_sha1_ascii)) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Downloaded config file did not match intended SHA1\n");
+		free(config_buf);
 		return -EINVAL;
 	}
 
-	return vpninfo->write_new_config(vpninfo, buf, buflen);
+	result = vpninfo->write_new_config(vpninfo, config_buf, buflen);
+	free(config_buf);
+	return result;
 }
 
 static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int buflen)
 {
 	char fname[16];
-	int fd;
+	int fd, ret;
 
 	if (!vpninfo->uid_csd_given) {
 		vpninfo->progress(vpninfo, PRG_ERR, "Error: You are trying to "
@@ -346,7 +384,13 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 				  strerror(errno));
 		return err;
 	}
-	write(fd, buf, buflen);
+
+	ret = proxy_write(fd, (void *)buf, buflen);
+	if (ret) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Failed to write temporary CSD script file: %s\n",
+				  strerror(ret));
+		return ret;
+	}
 	fchmod(fd, 0755);
 	close(fd);
 
@@ -372,7 +416,11 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 				exit(1);
 			}
 			setenv("HOME", pw->pw_dir, 1);
-			chdir(pw->pw_dir);
+			if (chdir(pw->pw_dir)) {
+				fprintf(stderr, "Failed to change to CSD home directory '%s': %s\n",
+					pw->pw_dir, strerror(errno));
+				exit(1);
+			}
 		}
 		if (vpninfo->uid_csd == 0) {
 			fprintf(stderr, "Warning: you are running insecure "
@@ -382,11 +430,14 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 
 		csd_argv[i++] = fname;
 		csd_argv[i++] = "-ticket";
-		asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket);
+		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
+			return -ENOMEM;
 		csd_argv[i++] = "-stub";
 		csd_argv[i++] = "\"0\"";
 		csd_argv[i++] = "-group";
-		asprintf(&csd_argv[i++], "\"%s\"", vpninfo->authgroup?:"");
+		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->authgroup?:"") == -1)
+			return -ENOMEM;
+
 		get_cert_md5_fingerprint(vpninfo, scert, scertbuf);
 		if (ccert)
 			get_cert_md5_fingerprint(vpninfo, ccert, ccertbuf);
@@ -394,16 +445,20 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 			ccertbuf[0] = 0;
 
 		csd_argv[i++] = "-certhash";
-		asprintf(&csd_argv[i++], "\"%s:%s\"", scertbuf, ccertbuf);
+		if (asprintf(&csd_argv[i++], "\"%s:%s\"", scertbuf, ccertbuf) == -1)
+			return -ENOMEM;
 		csd_argv[i++] = "-url";
-		asprintf(&csd_argv[i++], "\"https://%s%s\"", vpninfo->hostname, vpninfo->csd_starturl);
+		if (asprintf(&csd_argv[i++], "\"https://%s%s\"", vpninfo->hostname, vpninfo->csd_starturl) == -1)
+			return -ENOMEM;
 		/* WTF would it want to know this for? */
 		csd_argv[i++] = "-vpnclient";
 		csd_argv[i++] = "\"/opt/cisco/vpn/bin/vpnui";
 		csd_argv[i++] = "-connect";
-		asprintf(&csd_argv[i++], "https://%s/%s", vpninfo->hostname, vpninfo->csd_preurl);
+		if (asprintf(&csd_argv[i++], "https://%s/%s", vpninfo->hostname, vpninfo->csd_preurl) == -1)
+			return -ENOMEM;
 		csd_argv[i++] = "-connectparam";
-		asprintf(&csd_argv[i++], "#csdtoken=%s\"", vpninfo->csd_token);
+		if (asprintf(&csd_argv[i++], "#csdtoken=%s\"", vpninfo->csd_token) == -1)
+			return -ENOMEM;
 		csd_argv[i++] = "-langselen";
 		csd_argv[i++] = NULL;
 
@@ -445,6 +500,66 @@ char *local_strcasestr(const char *haystack, const char *needle)
 #define strcasestr local_strcasestr
 #endif
 
+int parse_url(char *url, char **res_proto, char **res_host, int *res_port,
+	      char **res_path, int default_port)
+{
+	char *proto = url;
+	char *host, *path, *port_str;
+	int port;
+
+	host = strstr(url, "://");
+	if (host) {
+		*host = 0;
+		host += 3;
+
+		if (!strcasecmp(proto, "https"))
+			port = 443;
+		else if (!strcasecmp(proto, "http"))
+			port = 80;
+		else if (!strcasecmp(proto, "socks") ||
+			 !strcasecmp(proto, "socks4") ||
+			 !strcasecmp(proto, "socks5"))
+			port = 1080;
+		else
+			return -EPROTONOSUPPORT;
+	} else {
+		if (default_port) {
+			proto = NULL;
+			port = default_port;
+			host = url;
+		} else
+			return -EINVAL;
+	}
+
+	path = strchr(host, '/');
+	if (path) {
+		*(path++) = 0;
+		if (!*path)
+			path = NULL;
+	}
+
+	port_str = strrchr(host, ':');
+	if (port_str) {
+		char *end;
+		int new_port = strtol(port_str + 1, &end, 10);
+
+		if (!*end) {
+			*port_str = 0;
+			port = new_port;
+		}
+	}
+
+	if (res_proto)
+		*res_proto = proto ? strdup(proto) : NULL;
+	if (res_host)
+		*res_host = strdup(host);
+	if (res_port)
+		*res_port = port;
+	if (res_path)
+		*res_path = path ? strdup(path) : NULL;
+	return 0;
+}
+
 /* Return value:
  *  < 0, on error
  *  = 0, no cookie (user cancel)
@@ -454,6 +569,7 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 {
 	struct vpn_option *opt, *next;
 	char buf[MAX_BUF_LEN];
+	char *form_buf = NULL;
 	int result, buflen;
 	char request_body[2048];
 	char *request_body_type = NULL;
@@ -497,12 +613,18 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	if (request_body_type)
 		sprintf(buf + strlen(buf), "%s", request_body);
 
-	vpninfo->progress(vpninfo, PRG_INFO, "%s %s/%s\n", method,
-			  vpninfo->hostname, vpninfo->urlpath ?: "");
+	if (vpninfo->port == 443)
+		vpninfo->progress(vpninfo, PRG_INFO, "%s https://%s/%s\n",
+				  method, vpninfo->hostname,
+				  vpninfo->urlpath ?: "");
+	else
+		vpninfo->progress(vpninfo, PRG_INFO, "%s https://%s:%d/%s\n",
+				  method, vpninfo->hostname, vpninfo->port,
+				  vpninfo->urlpath ?: "");
 
 	SSL_write(vpninfo->https_ssl, buf, strlen(buf));
 
-	buflen = process_http_response(vpninfo, &result, NULL, buf, MAX_BUF_LEN);
+	buflen = process_http_response(vpninfo, &result, NULL, &form_buf);
 	if (buflen < 0) {
 		/* We'll already have complained about whatever offended us */
 		exit(1);
@@ -512,27 +634,36 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 	redirect:
 		if (!strncmp(vpninfo->redirect_url, "https://", 8)) {
 			/* New host. Tear down the existing connection and make a new one */
-			char *host = vpninfo->redirect_url + 8;
-			char *path = strchr(host, '/');
+			char *host;
+			int port;
+			int ret;
 
 			free(vpninfo->urlpath);
-			if (path) {
-				*(path++) = 0;
-				vpninfo->urlpath = strdup(path);
-			} else
-				vpninfo->urlpath = NULL;
+			vpninfo->urlpath = NULL;
 
-			if (strcmp(vpninfo->hostname, host)) {
+			ret = parse_url(vpninfo->redirect_url, NULL, &host, &port, &vpninfo->urlpath, 0);
+			if (ret) {
+				vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse redirected URL '%s': %s\n",
+						  vpninfo->redirect_url, strerror(-ret));
+				free(vpninfo->redirect_url);
+				free(form_buf);
+				return ret;
+			}
+
+			if (strcasecmp(vpninfo->hostname, host) || port != vpninfo->port) {
 				free(vpninfo->hostname);
-				vpninfo->hostname = strdup(host);
+				vpninfo->hostname = host;
+				vpninfo->port = port;
 
 				/* Kill the existing connection, and a new one will happen */
 				free(vpninfo->peer_addr);
 				vpninfo->peer_addr = NULL;
-				SSL_free(vpninfo->https_ssl);
-				vpninfo->https_ssl = NULL;
-				close(vpninfo->ssl_fd);
-				vpninfo->ssl_fd = -1;
+				if (vpninfo->https_ssl) {
+					SSL_free(vpninfo->https_ssl);
+					vpninfo->https_ssl = NULL;
+					close(vpninfo->ssl_fd);
+					vpninfo->ssl_fd = -1;
+				}
 
 				for (opt = vpninfo->cookies; opt; opt = next) {
 					next = opt->next;
@@ -542,7 +673,9 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 					free(opt);
 				}
 				vpninfo->cookies = NULL;
-			}
+			} else
+				free(host);
+
 			free(vpninfo->redirect_url);
 			vpninfo->redirect_url = NULL;
 
@@ -555,40 +688,66 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			vpninfo->redirect_url = NULL;
 			goto retry;
 		} else {
-			vpninfo->progress(vpninfo, PRG_ERR, "Relative redirect (to '%s') not supported\n",
-				vpninfo->redirect_url);
-			return -EINVAL;
+			char *lastslash = strrchr(vpninfo->urlpath, '/');
+			if (!lastslash) {
+				free(vpninfo->urlpath);
+				vpninfo->urlpath = vpninfo->redirect_url;
+				vpninfo->redirect_url = NULL;
+			} else {
+				char *oldurl = vpninfo->urlpath;
+				*lastslash = 0;
+				vpninfo->urlpath = NULL;
+				if (asprintf(&vpninfo->urlpath, "%s/%s",
+					     oldurl, vpninfo->redirect_url) == -1) {
+					int err = -errno;
+					vpninfo->progress(vpninfo, PRG_ERR,
+							  "Allocating new path for relative redirect failed: %s\n",
+							  strerror(-err));
+					return err;
+				}
+				free(oldurl);
+				free(vpninfo->redirect_url);
+				vpninfo->redirect_url = NULL;
+			}
+			goto retry;
 		}
 	}
 
 	if (vpninfo->csd_stuburl) {
 		/* This is the CSD stub script, which we now need to run */
-		result = run_csd_script(vpninfo, buf, buflen);
-		if (result)
+		result = run_csd_script(vpninfo, form_buf, buflen);
+		if (result) {
+			free(form_buf);
 			return result;
+		}
 
 		/* Now we'll be redirected to the waiturl */
 		goto retry;
 	}
-	if (strncmp(buf, "<?xml", 5)) {
+	if (strncmp(form_buf, "<?xml", 5)) {
 		/* Not XML? Perhaps it's HTML with a refresh... */
-		if (strcasestr(buf, "http-equiv=\"refresh\"")) {
+		if (strcasestr(form_buf, "http-equiv=\"refresh\"")) {
 			vpninfo->progress(vpninfo, PRG_INFO, "Refreshing %s after 1 second...\n",
 					  vpninfo->urlpath);
 			sleep(1);
 			goto retry;
 		}
 		vpninfo->progress(vpninfo, PRG_ERR, "Unknown response from server\n");
+		free(form_buf);
 		return -EINVAL;
 	}
 	request_body[0] = 0;
-	result = parse_xml_response(vpninfo, buf, request_body, sizeof(request_body),
+	result = parse_xml_response(vpninfo, form_buf, request_body, sizeof(request_body),
 				    &method, &request_body_type);
+
 	if (!result)
 		goto redirect;
 
+	free(form_buf);
+
 	if (result != 2)
 		return result;
+
 	/* A return value of 2 means the XML form indicated
 	   success. We _should_ have a cookie... */
 
@@ -637,3 +796,281 @@ char *openconnect_create_useragent(char *base)
 
 	return uagent;
 }
+
+static int proxy_gets(int fd, char *buf, size_t len)
+{
+	int i = 0;
+	int ret;
+
+	if (len < 2)
+		return -EINVAL;
+
+	while ( (ret = read(fd, buf + i, 1)) == 1) {
+		if (buf[i] == '\n') {
+			buf[i] = 0;
+			if (i && buf[i-1] == '\r') {
+				buf[i-1] = 0;
+				i--;
+			}
+			return i;
+		}
+		i++;
+
+		if (i >= len - 1) {
+			buf[i] = 0;
+			return i;
+		}
+	}
+	if (ret < 0)
+		ret = -errno;
+
+	buf[i] = 0;
+	return i ?: ret;
+}
+
+static int proxy_write(int fd, unsigned char *buf, size_t len)
+{
+	size_t count;
+	
+	for (count = 0; count < len; ) {
+		int i = write(fd, buf + count, len - count);
+		if (i < 0)
+			return -errno;
+
+		count += i;
+	}
+	return 0;
+}
+
+static int proxy_read(int fd, unsigned char *buf, size_t len)
+{
+	size_t count;
+
+	for (count = 0; count < len; ) {
+		int i = read(fd, buf + count, len - count);
+		if (i < 0)
+			return -errno;
+
+		count += i;
+	}
+	return 0;
+}
+
+static const char *socks_errors[] = {
+	"request granted",
+	"general failure",
+	"connection not allowed by ruleset",
+	"network unreachable",
+	"host unreachable",
+	"connection refused by destination host",
+	"TTL expired",
+	"command not supported / protocol error",
+	"address type not supported"
+};
+
+static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
+{
+	unsigned char buf[1024];
+	int i;
+
+	buf[0] = 5; /* SOCKS version */
+	buf[1] = 1; /* # auth methods */
+	buf[2] = 0; /* No auth supported */
+
+	if ((i = proxy_write(ssl_sock, buf, 3))) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error writing auth request to SOCKS proxy: %s\n",
+				  strerror(-i));
+		return i;
+	}
+	
+	if ((i = proxy_read(ssl_sock, buf, 2))) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error reading auth response from SOCKS proxy: %s\n",
+				  strerror(-i));
+		return i;
+	}
+	if (buf[0] != 5) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unexpected auth response from SOCKS proxy: %02x %02x\n",
+				  buf[0], buf[1]);
+		return -EIO;
+	}
+	if (buf[1]) {
+	socks_err:
+		if (buf[1] < sizeof(socks_errors) / sizeof(socks_errors[0]))
+			vpninfo->progress(vpninfo, PRG_ERR,
+					  "SOCKS proxy error %02x: %s\n",
+					  buf[1], socks_errors[buf[1]]);
+		else
+			vpninfo->progress(vpninfo, PRG_ERR,
+					  "SOCKS proxy error %02x\n",
+					  buf[1]);
+		return -EIO;
+	}
+
+	vpninfo->progress(vpninfo, PRG_INFO, "Requesting SOCKS proxy connection to %s:%d\n",
+			  vpninfo->hostname, vpninfo->port);
+
+	buf[0] = 5; /* SOCKS version */
+	buf[1] = 1; /* CONNECT */
+	buf[2] = 0; /* Reserved */
+	buf[3] = 3; /* Address type is domain name */
+	buf[4] = strlen(vpninfo->hostname);
+	strcpy((char *)buf + 5, vpninfo->hostname);
+	i = strlen(vpninfo->hostname) + 5;
+	buf[i++] = vpninfo->port >> 8;
+	buf[i++] = vpninfo->port & 0xff;
+
+	if ((i = proxy_write(ssl_sock, buf, i))) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error writing connect request to SOCKS proxy: %s\n",
+				  strerror(-i));
+		return i;
+	}
+	/* Read 5 bytes -- up to and including the first byte of the returned
+	   address (which might be the length byte of a domain name) */
+	if ((i = proxy_read(ssl_sock, buf, 5))) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error reading connect response from SOCKS proxy: %s\n",
+				  strerror(-i));
+		return i;
+	}
+	if (buf[0] != 5) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unexpected connect response from SOCKS proxy: %02x %02x...\n",
+				  buf[0], buf[1]);
+		return -EIO;
+	}
+	if (buf[1])
+		goto socks_err;
+
+	/* Connect responses contain an address */
+	switch(buf[3]) {
+	case 1: /* Legacy IP */
+		i = 5;
+		break;
+	case 3: /* Domain name */
+		i = buf[4] + 2;
+		break;
+	case 4: /* IPv6 */
+		i = 17;
+		break;
+	default:
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unexpected address type %02x in SOCKS connect response\n",
+				  buf[3]);
+		return -EIO;
+	}
+
+	if ((i = proxy_read(ssl_sock, buf, i))) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Error reading connect response from SOCKS proxy: %s\n",
+				  strerror(-i));
+		return i;
+	}
+	return 0;
+}
+
+static int process_http_proxy(struct openconnect_info *vpninfo, int ssl_sock)
+{
+	char buf[MAX_BUF_LEN];
+	int buflen, result;
+
+	sprintf(buf, "CONNECT %s:%d HTTP/1.1\r\n", vpninfo->hostname, vpninfo->port);
+	sprintf(buf + strlen(buf), "Host: %s\r\n", vpninfo->hostname);
+	sprintf(buf + strlen(buf), "User-Agent: %s\r\n", vpninfo->useragent);
+	sprintf(buf + strlen(buf), "Proxy-Connection: keep-alive\r\n");
+	sprintf(buf + strlen(buf), "Connection: keep-alive\r\n");
+	sprintf(buf + strlen(buf), "Accept-Encoding: identity\r\n");
+	sprintf(buf + strlen(buf), "\r\n");
+
+	vpninfo->progress(vpninfo, PRG_INFO, "Requesting HTTP proxy connection to %s:%d\n",
+			  vpninfo->hostname, vpninfo->port);
+
+	if (proxy_write(ssl_sock, (unsigned char *)buf, strlen(buf))) {
+		result = -errno;
+		vpninfo->progress(vpninfo, PRG_ERR, "Sending proxy request failed: %s\n",
+				  strerror(errno));
+		return result;
+	}
+
+	if (proxy_gets(ssl_sock, buf, sizeof(buf)) < 0) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Error fetching proxy response\n");
+		return -EIO;
+	}
+
+	if (strncmp(buf, "HTTP/1.", 7) || (buf[7] != '0' && buf[7] != '1') ||
+	    buf[8] != ' ' || !(result = atoi(buf+9))) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Failed to parse proxy response '%s'\n",
+				  buf);
+		return -EINVAL;
+	}
+
+	if (result != 200) {
+		vpninfo->progress(vpninfo, PRG_ERR, "Proxy CONNECT request failed: %s\n",
+				  buf);
+		return -EIO;
+	}
+
+	while ((buflen = proxy_gets(ssl_sock, buf, sizeof(buf)))) {
+		if (buflen < 0) {
+			vpninfo->progress(vpninfo, PRG_ERR, "Failed to read proxy response\n");
+			return -EIO;
+		}
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Unexpected continuation line after CONNECT response: '%s'\n",
+				  buf);
+	}
+
+	return 0;
+}
+
+int process_proxy(struct openconnect_info *vpninfo, int ssl_sock)
+{
+	if (!vpninfo->proxy_type || !strcmp(vpninfo->proxy_type, "http"))
+		return process_http_proxy(vpninfo, ssl_sock);
+	
+	if (!strcmp(vpninfo->proxy_type, "socks") ||
+	    !strcmp(vpninfo->proxy_type, "socks5"))
+		return process_socks_proxy(vpninfo, ssl_sock);
+
+	vpninfo->progress(vpninfo, PRG_ERR, "Unknown proxy type '%s'\n",
+				  vpninfo->proxy_type);
+	return -EIO;
+}
+
+int set_http_proxy(struct openconnect_info *vpninfo, char *proxy)
+{
+	char *url = strdup(proxy);
+	int ret;
+
+	if (!url)
+		return -ENOMEM;
+
+	free(vpninfo->proxy_type);
+	vpninfo->proxy_type = NULL;
+	free(vpninfo->proxy);
+	vpninfo->proxy = NULL;
+
+	ret = parse_url(url, &vpninfo->proxy_type, &vpninfo->proxy,
+			&vpninfo->proxy_port, NULL, 80);
+	if (ret)
+		goto out;
+
+	if (vpninfo->proxy_type &&
+	    strcmp(vpninfo->proxy_type, "http") &&
+	    strcmp(vpninfo->proxy_type, "socks") &&
+	    strcmp(vpninfo->proxy_type, "socks5")) {
+		vpninfo->progress(vpninfo, PRG_ERR,
+				  "Only http or socks(5) proxies supported\n");
+		free(vpninfo->proxy_type);
+		vpninfo->proxy_type = NULL;
+		free(vpninfo->proxy);
+		vpninfo->proxy = NULL;
+		return -EINVAL;
+	}
+ out:
+	free(url);
+	return ret;
+}
