@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  *
  * Authors: Jussi Kukkonen <jku@linux.intel.com>
  *          David Woodhouse <dwmw2@infradead.org>
@@ -504,6 +504,14 @@ static gboolean ui_form (struct oc_auth_form *form)
 
 	ssl_box_clear(ui_data);
 
+	g_mutex_lock(ui_data->form_mutex);
+	while (!g_queue_is_empty (ui_data->form_entries)) {
+		ui_fragment_data *data;
+		data = g_queue_pop_tail (ui_data->form_entries);
+		g_slice_free (ui_fragment_data, data);
+	}
+	g_mutex_unlock(ui_data->form_mutex);
+
 	if (form->banner)
 		ssl_box_add_info(ui_data, form->banner);
 	if (form->error)
@@ -1114,8 +1122,10 @@ static gboolean cookie_obtained(auth_ui_data *ui_data)
 			g_free(k);
 		}
 
-		printf("%s\n%s\n", NM_OPENCONNECT_KEY_GATEWAY, ui_data->vpninfo->hostname);
-		printf("%s\n%s\n", NM_OPENCONNECT_KEY_COOKIE, ui_data->vpninfo->cookie);
+		printf("%s\n%s:%d\n", NM_OPENCONNECT_KEY_GATEWAY,
+		       ui_data->vpninfo->hostname, ui_data->vpninfo->port);
+		printf("%s\n%s\n", NM_OPENCONNECT_KEY_COOKIE,
+		       ui_data->vpninfo->cookie);
 		print_peer_cert(ui_data->vpninfo);
 		memset((void *)ui_data->vpninfo->cookie, 0, strlen(ui_data->vpninfo->cookie));
 		printf("\n\n");
@@ -1185,8 +1195,15 @@ static void connect_host(auth_ui_data *ui_data)
 	for (i = 0; i < host_nr; i++)
 		host = host->next;
 
-	ui_data->vpninfo->hostname = g_strdup(host->hostaddress);
-	if (host->usergroup)
+	if (parse_url(host->hostaddress, NULL,
+		      &ui_data->vpninfo->hostname, &ui_data->vpninfo->port,
+		      &ui_data->vpninfo->urlpath, 443)) {
+		fprintf(stderr, "Failed to parse server URL '%s'\n",
+			host->hostaddress);
+		ui_data->vpninfo->hostname = g_strdup(host->hostaddress);
+	}
+
+	if (!ui_data->vpninfo->urlpath && host->usergroup)
 		ui_data->vpninfo->urlpath = g_strdup(host->usergroup);
 
 	remember_gconf_key(ui_data, g_strdup("lasthost"), g_strdup(host->hostname));
@@ -1386,6 +1403,9 @@ static auth_ui_data *init_ui_data (char *vpn_name)
 	ui_data->vpninfo->validate_peer_cert = validate_peer_cert;
 	ui_data->vpninfo->vpn_name = vpn_name;
 	ui_data->vpninfo->process_auth_form = nm_process_auth_form;
+#if 0
+	ui_data->vpninfo->proxy_factory = px_proxy_factory_new();
+#endif
 
 	return ui_data;
 }
