@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -34,6 +34,10 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <unistd.h>
+#ifdef OPENCONNECT_LIBPROXY
+#include <libproxy/proxy.h>
+#endif
+
 
 /****************************************************************************/
 
@@ -143,8 +147,16 @@ struct openconnect_info {
 	char sid_tokencode[9];
 	char sid_nexttokencode[9];
 
+#ifdef OPENCONNECT_LIBPROXY
+	pxProxyFactory *proxy_factory;
+#endif
+	char *proxy_type;
+	char *proxy;
+	int proxy_port;
+
 	const char *localname;
 	char *hostname;
+	int port;
 	char *urlpath;
 	const char *cert;
 	const char *sslkey;
@@ -230,6 +242,7 @@ struct openconnect_info {
 
 	socklen_t peer_addrlen;
 	struct sockaddr *peer_addr;
+	struct sockaddr *dtls_addr;
 
 	int deflate;
 	char *useragent;
@@ -319,6 +332,10 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 /* http.c */
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo);
 char *openconnect_create_useragent(char *base);
+int process_proxy(struct openconnect_info *vpninfo, int ssl_sock);
+int parse_url(char *url, char **res_proto, char **res_host, int *res_port,
+	      char **res_path, int default_port);
+int set_http_proxy(struct openconnect_info *vpninfo, char *proxy);
 
 /* ssl_ui.c */
 int set_openssl_ui(void);
