@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -49,6 +49,11 @@
 
 #include "openconnect.h"
 
+/* OSX < 1.6 doesn't have AI_NUMERICSERV */
+#ifndef AI_NUMERICSERV
+#define AI_NUMERICSERV 0
+#endif
+
 /* Helper functions for reading/writing lines over SSL.
    We could use cURL for the HTTP stuff, but it's overkill */
 
@@ -398,7 +403,7 @@ static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
 	if (ret)
 		return ret;
 
-	if (strcmp(vpninfo->servercert, fingerprint)) {
+	if (strcasecmp(vpninfo->servercert, fingerprint)) {
 		vpninfo->progress(vpninfo, PRG_ERR,
 				  "Server SSL certificate didn't match: %s\n", fingerprint);
 		return -EINVAL;
@@ -477,20 +482,25 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	int ssl_sock = -1;
 	int err;
 
+	if (!vpninfo->port)
+		vpninfo->port = 443;
+
 	if (vpninfo->peer_addr) {
 		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
 		if (ssl_sock < 0) {
 		reconn_err:
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to reconnect to host %s\n", vpninfo->hostname);
+			vpninfo->progress(vpninfo, PRG_ERR, "Failed to reconnect to %s %s\n",
+					  vpninfo->proxy?"proxy":"host",
+					  vpninfo->proxy?:vpninfo->hostname);
 			return -EINVAL;
 		}
 		if (connect(ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen))
 			goto reconn_err;
-
-		
 		
 	} else {
 		struct addrinfo hints, *result, *rp;
+		char *hostname;
+		char port[6];
 
 		memset(&hints, 0, sizeof(struct addrinfo));
 		hints.ai_family = AF_UNSPEC;
@@ -501,16 +511,93 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		hints.ai_addr = NULL;
 		hints.ai_next = NULL;
 
-		err = getaddrinfo(vpninfo->hostname, "443", &hints, &result);
+		/* The 'port' variable is a string because it's easier
+		   this way than if we pass NULL to getaddrinfo() and
+		   then try to fill in the numeric value into
+		   different types of returned sockaddr_in{6,}. */
+#ifdef OPENCONNECT_LIBPROXY
+		if (vpninfo->proxy_factory) {
+			char *url;
+			char **proxies;
+			int i = 0;
+
+			free(vpninfo->proxy_type);
+			vpninfo->proxy_type = NULL;
+			free(vpninfo->proxy);
+			vpninfo->proxy = NULL;
+
+			if (vpninfo->port == 443)
+				i = asprintf(&url, "https://%s/%s", vpninfo->hostname,
+					     vpninfo->urlpath?:"");
+			else
+				i = asprintf(&url, "https://%s:%d/%s", vpninfo->hostname,
+					     vpninfo->port, vpninfo->urlpath?:"");
+			if (i == -1)
+				return -ENOMEM;
+
+			proxies = px_proxy_factory_get_proxies(vpninfo->proxy_factory,
+							       url);
+
+			while (proxies && proxies[i]) {
+				if (!vpninfo->proxy &&
+				    (!strncmp(proxies[i], "http://", 7) ||
+				     !strncmp(proxies[i], "socks://", 8) ||
+				     !strncmp(proxies[i], "socks5://", 9)))
+					parse_url(proxies[i], &vpninfo->proxy_type,
+						  &vpninfo->proxy, &vpninfo->proxy_port,
+						  NULL, 0);
+				i++;
+			}
+			free(url);
+			free(proxies);
+			if (vpninfo->proxy)
+				vpninfo->progress(vpninfo, PRG_TRACE, "Proxy from libproxy: %s://%s:%d/\n",
+						  vpninfo->proxy_type, vpninfo->proxy, vpninfo->port);
+		}
+#endif
+		if (vpninfo->proxy) {
+			hostname = vpninfo->proxy;
+			snprintf(port, 6, "%d", vpninfo->proxy_port);
+		} else {
+			hostname = vpninfo->hostname;
+			snprintf(port, 6, "%d", vpninfo->port);
+		}
+
+		if (hostname[0] == '[' && hostname[strlen(hostname)-1] == ']') {
+			/* Solaris has no strndup(). */
+			int len = strlen(hostname) - 2;
+			char *new_hostname = malloc(len + 1);
+			if (!new_hostname)
+				return -ENOMEM;
+			memcpy(new_hostname, hostname + 1, len);
+			new_hostname[len] = 0;
+
+			hostname = new_hostname;
+			hints.ai_flags |= AI_NUMERICHOST;
+		}
+
+		err = getaddrinfo(hostname, port, &hints, &result);
+		if (hints.ai_flags & AI_NUMERICHOST)
+			free(hostname);
+
 		if (err) {
-			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed: %s\n", gai_strerror(err));
+			vpninfo->progress(vpninfo, PRG_ERR, "getaddrinfo failed: %s\n",
+					  gai_strerror(err));
 			return -EINVAL;
 		}
 
-		vpninfo->progress(vpninfo, PRG_INFO,
-				  "Attempting to connect to %s\n", vpninfo->hostname);
-
 		for (rp = result; rp ; rp = rp->ai_next) {
+			char host[80];
+
+			if (!getnameinfo(rp->ai_addr, rp->ai_addrlen, host,
+					 sizeof(host), NULL, 0, NI_NUMERICHOST))
+				vpninfo->progress(vpninfo, PRG_INFO,
+						  "Attempting to connect to %s%s%s:%s\n",
+						  rp->ai_family == AF_INET6?"[":"",
+						  host,
+						  rp->ai_family == AF_INET6?"]":"",
+						  port);
+			
 			ssl_sock = socket(rp->ai_family, rp->ai_socktype,
 					  rp->ai_protocol);
 			if (ssl_sock < 0)
@@ -534,12 +621,21 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		freeaddrinfo(result);
 		
 		if (ssl_sock < 0) {
-			vpninfo->progress(vpninfo, PRG_ERR, "Failed to connect to host %s\n", vpninfo->hostname);
+			vpninfo->progress(vpninfo, PRG_ERR, "Failed to connect to host %s\n",
+					  vpninfo->proxy?:vpninfo->hostname);
 			return -EINVAL;
 		}
 	}
 	fcntl(ssl_sock, F_SETFD, FD_CLOEXEC);
 
+	if (vpninfo->proxy) {
+		err = process_proxy(vpninfo, ssl_sock);
+		if (err) {
+			close(ssl_sock);
+			return err;
+		}
+	}
+
 	ssl3_method = TLSv1_client_method();
 	if (!vpninfo->https_ctx) {
 		vpninfo->https_ctx = SSL_CTX_new(ssl3_method);
