@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008 Intel Corporation.
+ * Copyright © 2008-2010 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -30,6 +30,7 @@
 #include <signal.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <netdb.h>
 #include <netinet/in_systm.h>
 #include <netinet/in.h>
 #include <netinet/ip.h>
@@ -37,7 +38,6 @@
 #include <arpa/inet.h>
 #include <errno.h>
 #if defined(__sun__)
-#include <net/if_tun.h>
 #include <stropts.h>
 #include <sys/sockio.h>
 #endif
@@ -234,10 +234,12 @@ static void setenv_cstp_opts(struct openconnect_info *vpninfo)
 
 static void set_script_env(struct openconnect_info *vpninfo)
 {
-	struct sockaddr_in *sin = (void *)vpninfo->peer_addr;
+	char host[80];
+	int ret = getnameinfo(vpninfo->peer_addr, vpninfo->peer_addrlen, host,
+			      sizeof(host), NULL, 0, NI_NUMERICHOST);
+	if (!ret)
+		setenv("VPNGATEWAY", host, 1);
 
-	setenv("VPNGATEWAY", inet_ntoa(sin->sin_addr), 1);
-	setenv("TUNDEV", vpninfo->ifname, 1);
 	setenv("reason", "connect", 1);
 	unsetenv("CISCO_BANNER");
 	unsetenv("CISCO_SPLIT_INC");
@@ -331,15 +333,13 @@ static void set_script_env(struct openconnect_info *vpninfo)
 
 static int script_config_tun(struct openconnect_info *vpninfo)
 {
-	if (vpninfo->peer_addr->sa_family != AF_INET || !vpninfo->vpn_addr) {
+	if (system(vpninfo->vpnc_script)) {
+		int e = errno;
 		vpninfo->progress(vpninfo, PRG_ERR,
-				  "Script can only handle Legacy IP\n");
-		return -EINVAL;
+				  "Failed to spawn script '%s': %s\n",
+				  vpninfo->vpnc_script, strerror(e));
+		return -e;
 	}
-
-	set_script_env(vpninfo);
-
-	system(vpninfo->vpnc_script);
 	return 0;
 }
 
@@ -349,6 +349,8 @@ int setup_tun(struct openconnect_info *vpninfo)
 {
 	int tun_fd;
 
+	set_script_env(vpninfo);
+
 	if (vpninfo->script_tun) {
 		pid_t child;
 		int fds[2];
@@ -495,6 +497,7 @@ int setup_tun(struct openconnect_info *vpninfo)
 #endif
 #endif
 		if (vpninfo->vpnc_script) {
+			setenv("TUNDEV", vpninfo->ifname, 1);
 			script_config_tun(vpninfo);
 			/* We have to set the MTU for ourselves, because the script doesn't */
 			local_config_tun(vpninfo, 1);
@@ -594,9 +597,13 @@ void shutdown_tun(struct openconnect_info *vpninfo)
 		kill(vpninfo->script_tun, SIGHUP);
 	} else {
 		if (vpninfo->vpnc_script) {
-			setenv("TUNDEV", vpninfo->ifname, 1);
 			setenv("reason", "disconnect", 1);
-			system(vpninfo->vpnc_script);
+			if (system(vpninfo->vpnc_script) == -1) {
+				vpninfo->progress(vpninfo, PRG_ERR,
+						  "Failed to spawn script '%s': %s\n",
+						  vpninfo->vpnc_script,
+						  strerror(errno));
+			}
 		}
 #ifdef __sun__
 		if (ioctl(vpninfo->ip_fd, I_PUNLINK, vpninfo->tun_muxid) < 0)
